#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <windows.h>
#include <psapi.h>

#include "omp.h"
#include "mkl.h"
#include "mkl_lapack.h"

void init_symm_matrix(const int& msize, double* a)
{
	// This builds a symmetric matrix with dimensions msize x msize and
	// populated with 1.0 on the diagonal and 0.5 on the off diagonal terms.

	const double p = 1.0;

	for (auto i = 0; i < msize; ++i)
	{
		for (auto j = i; j < msize; ++j)
		{
			// Fill the diagonal elements

			if (i == j)
			{
				*(a + i * msize + j) = p;
			}
			else
			{
				// fill upper triangular part
				*(a + i * msize + j) = p / 2;

				// fill lower triangular part
				*(a + j * msize + i) = p / 2;
			}

		}
	}
}


void symm_invert(const int N, const double *matrix_in, double *result_matrix, int& ier)
{
	// copy the input matrix to the result

	for (auto i = 0; i < N; ++i)
	{
		for (auto j = 0; j < N; ++j)
		{
			*(result_matrix + i * N + j) = *(matrix_in + i * N + j);
		}
	}

	
	ier = 0;
	const char uplo = 'U';
	dpotrf(&uplo, &N, result_matrix, &N, &ier);

	if (true)
	{
		if (ier == 0)
		{
			dpotri(&uplo, &N, result_matrix, &N, &ier);

			// we normally run this code if we care about getting the actual inverse matrix back.
			// for this test we don't.

			if (false)
			// Copy from lower to upper triangle

			for (int i = 0; i < N; ++i)
			{
				double* uptr = result_matrix + i * N + i + 1;
				double* lptr = uptr + N - 1;

				for (int j = i + 1; j < N; ++j)
				{
					*uptr = *lptr;
					++uptr;		// move across row
					lptr += N;	// move down column
				}
			}
		}
	}
}

void parse_command_line(const int argc, char *argv[], int &msize, int&num_threads, bool &use_mkl)
{
	// parse the command line arguments [matrix size] [number of threads] [threading control OMP | MKL]

	if (argc >= 2)
	{
		msize = atoi(argv[1]);
	}

	if (argc >= 3)
	{
		num_threads = atoi(argv[2]);
	}

	if (argc == 4)
	{
		use_mkl = !strcmp("MKL", argv[3]);
	}
}

void print_MKL_version()
{
	const int len = 198;
	char version[len];
	mkl_get_version_string(version, len);

	printf("Using MKL version %s\n", version);
}

void print_env_variables()
{
	// Check if any of the threading environment variables are set

	auto this_env = getenv("OMP_NUM_THREADS");
	if (this_env)
	{
		printf("Warning: OMP_NUM_THREADS is set to %s\n", this_env);
	}

	this_env = getenv("MKL_NUM_THREADS");
	if (this_env)
	{
		printf("Warning: MKL_NUM_THREADS is set to %s\n", this_env);
	}

	this_env = getenv("MKL_DOMAIN_NUM_THREADS");
	if (this_env)
	{
		printf("Warning: MKL_DOMAIN_NUM_THREADS is set to %s\n", this_env);
	}
}

void set_threading_control(int &num_threads, bool use_mkl)
{
	// What's our threading capabilities?

	const int max_omp_threads = omp_get_max_threads();
	printf("Maximum number of OMP threads: %d\n", max_omp_threads);
	const int max_mkl_threads = mkl_get_max_threads();
	printf("Maximum number of MKL threads: %d\n", max_mkl_threads);

	if (max_omp_threads != max_mkl_threads)
	{
		printf("Note: OMP threads != MKL threads - Do you have hyperthreading enabled?\n");
	}

	// by default use the maximum number of threads unless otherwise
	// specified.

	if (num_threads == 0)
	{
		if (use_mkl)
		{
			num_threads = max_mkl_threads;
		}
		else
		{
			num_threads = max_omp_threads;
		}
	}

	// If we set the MKL threads we will have a leak
	// If we set the OMP threads we don't have a leak

	if (use_mkl)
	{
		mkl_set_num_threads(num_threads);
	}
	else
	{
		omp_set_num_threads(num_threads);
	}

}
int main(int argc, char *argv[])
{

	printf("Matrix test started\n");

	// define default values for input arguments

	// matrix size <= 32 doesn't leak. Make it 33 or larger and we leak.
	int msize = 50;

	// by default use all available threads. If using 2 or more threads we see the leak.
	int num_threads = 0;

	// by default use OMP threads
	bool use_mkl = false;

	parse_command_line(argc, argv, msize, num_threads, use_mkl);
	print_MKL_version();
	print_env_variables();
	set_threading_control(num_threads, use_mkl);

	printf("Matrix size %d, Number of threads %d, Use MKL threads? %d\n", msize, num_threads, use_mkl);
	
	// This is the only memory allocation we do ... and once
	auto *a = static_cast<double*>(_aligned_malloc(sizeof(double) * msize * msize, 64));
	auto *b = static_cast<double*>(_aligned_malloc(sizeof(double) * msize * msize, 64));

	// initialize a to make it symmetric
	init_symm_matrix(msize, a);

	// This has no effect ...
	mkl_disable_fast_mm();

	// Try recording the memory usage.
	mkl_peak_mem_usage(MKL_PEAK_MEM_ENABLE);
	
	// testing inverting the matrix a bunch of times
	constexpr size_t test_count = 100000;
	for (size_t i = 0; i < test_count; ++i)
	{
		int ier = 0;
		symm_invert(msize, a, b, ier);

		// This has no effect ...
		mkl_thread_free_buffers();

		// Nor does this ...
		mkl_free_buffers();
	}

	_aligned_free(a);
	_aligned_free(b);

	// This is a very large number ...
	const auto bytes_used = mkl_peak_mem_usage(MKL_PEAK_MEM_RESET) / 1024;
	printf("   MKL peak mem usage [KBytes]: %lld\n", bytes_used);

	// Get the current memory size of the process to see if we are leaking?

	PROCESS_MEMORY_COUNTERS pmc;
	const auto pid = GetCurrentProcess();
	GetProcessMemoryInfo(pid, &pmc, sizeof(pmc));
	const size_t mem_size = pmc.WorkingSetSize / 1024;
	printf("Working set mem usage [Kbytes]: %llu\n", mem_size);
	if (mem_size > 10000)
	{
		printf("Working set > 10000 K, looks like you are leaking!\n");
	}
	else
	{
		printf("Working < 10000 K, no leak found!\n");
	}

	printf("Matrix test completed\n\n");
}

